"""Pages processing methods and functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

from typing import Dict, Any

from dfcx_scrapi.agent_extract import common
from dfcx_scrapi.agent_extract import types
from dfcx_scrapi.agent_extract import routes


class Pages:
    """Pages processing methods and functions."""

    def __init__(self):
        self.common = common.Common()
        self.routes = routes.Fulfillments()

    @staticmethod
    def build_page_path_list(flow_path: str):
        """Builds a list of files, each representing a Page.

        Ex: /path/to/agent/flows/<flow_dir>/pages/<page_name>.json
        """
        pages_path = f"{flow_path}/pages"

        page_paths = []

        for page in os.listdir(pages_path):
            page_file_path = f"{pages_path}/{page}"
            page_paths.append(page_file_path)

        return page_paths

    @staticmethod
    def get_form_parameter_data(param: Dict[str, Any], page: types.Page):
        fp = types.FormParameter(page=page)
        fp.display_name = param.get("displayName", None)
        fp.entity_type = param.get("entityType", None)
        fp.required = param.get("required", None)

        fp.fill_behavior = param.get("fillBehavior", None)

        if fp.fill_behavior:
            fp.init_fulfillment = fp.fill_behavior.get(
                "initialPromptFulfillment", None)
            fp.reprompt_handlers = fp.fill_behavior.get(
                "repromptEventHandlers", None)

        fp.advanced_settings = page.form.get("advancedSettings", None)

        if fp.advanced_settings:
            fp.dtmf_settings = fp.advanced_settings.get("dtmfSettings", None)

        return fp

    def process_form(self, page: types.Page, stats: types.AgentData):
        """Process the Form and sub-resources within it for the Page."""
        parameters = page.form.get("parameters", None)

        if parameters:
            for param in parameters:
                fp = self.get_form_parameter_data(param, page)
                stats = self.routes.process_reprompt_handlers(fp, stats)

        return stats


    def process_page(self, page: types.Page, stats: types.AgentData):
        """Process a Single Page file."""
        page.display_name = self.common.parse_filepath(page.page_file, "page")
        page.display_name = self.common.clean_display_name(page.display_name)

        stats.graph.add_node(page.display_name)
        page.flow.graph.add_node(page.display_name)

        page.flow.all_pages.add(page.display_name)

        with open(page.page_file, "r", encoding="UTF-8") as page_file:
            page.data = json.load(page_file)
            page.entry = page.data.get("entryFulfillment", None)
            page.events = page.data.get("eventHandlers", None)
            page.form = page.data.get("form", None)
            page.routes = page.data.get("transitionRoutes", None)
            page.route_groups = page.data.get("transitionRouteGroups", None)
            page.resource_id = page.data.get("name", None)

            # Order of linting is important here
            stats = self.routes.process_entry(page, stats)
            stats = self.routes.process_routes(page, stats)
            stats = self.routes.process_events(page, stats)
            stats = self.process_form(page, stats)

            if page.route_groups:
                page, stats = self.routes.set_route_group_targets(page, stats)

            page_file.close()

        full_flow_id = f"{stats.agent_id}/flows/{page.flow.resource_id}"
        full_page_id = f"{full_flow_id}/pages/{page.resource_id}"
        stats.pages[page.flow.display_name].append(page.data)
        stats.flow_page_map[
            page.flow.display_name]["pages"][page.display_name] = full_page_id

        return stats

    def process_pages_directory(self, flow: types.Flow, stats: types.AgentData):
        """Process the Pages dir inside a specific Flow dir.

        Some Flows may not contain Pages, so we check for the existence
        of the directory before traversing
        """
        if "pages" in os.listdir(flow.dir_path):
            page_paths = self.build_page_path_list(flow.dir_path)

            for page_path in page_paths:
                page = types.Page(flow=flow)
                page.agent_id = flow.agent_id
                page.page_file = page_path
                stats.total_pages += 1
                stats = self.process_page(page, stats)

        return stats



"""A set of Utility methods to check resources stats on DFCX Agents."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict
from dataclasses import dataclass

import logging
import datetime
import pandas as pd
import gspread

from dfcx_scrapi.core import scrapi_base
from dfcx_scrapi.core import agents
from dfcx_scrapi.core import flows
from dfcx_scrapi.core import pages
from dfcx_scrapi.core import intents
from dfcx_scrapi.core import conversation
from dfcx_scrapi.tools import dataframe_functions

pd.options.display.max_colwidth = 200

GLOBAL_SCOPE = [
    "https://spreadsheets.google.com/feeds",
    "https://www.googleapis.com/auth/drive",
]
INPUT_SCHEMA_COLUMNS = [
        "flow_display_name",
        "page_display_name",
        "utterance",
        "expected_intent",
        "expected_parameters",
        "description",
    ]

OUTPUT_SCHEMA_COLUMNS = [
    "flow_display_name",
    "page_display_name",
    "utterance",
    "expected_intent",
    "expected_parameters",
    "target_page",
    "match_type",
    "confidence",
    "parameters_set",
    "detected_intent",
    "agent_display_name",
    "description",
    "input_source"
    ]

SUMMARY_SCHEMA_COLUMNS = [
    "test_run_timestamp",
    "total_tests",
    "pass_count",
    "pass_rate",
    "no_match_count",
    "no_match_rate",
    "test_agent",
    "data_source"
    ]

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


@dataclass
class Stats:
    """Dataclass for the summary stats."""
    no_match_count: int = 0
    no_match_rate: float = 0.0
    pass_count: int = 0
    pass_rate: float = 0.0
    test_agent: str = None
    data_source: str = None


class NluEvals(scrapi_base.ScrapiBase):
    """NLU Evaluation Class for Dialogflow CX Testing."""
    def __init__(
        self,
        agent_id: str,
        creds_path: str = None,
        creds_dict: Dict[str, str] = None,
        creds=None,
    ):

        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=GLOBAL_SCOPE
        )

        self.agent_id = agent_id
        self._sheets_client = self._build_sheets_client()

        self._a = agents.Agents(creds=self.creds)
        self._i = intents.Intents(creds=self.creds)
        self._f = flows.Flows(creds=self.creds)
        self._p = pages.Pages(creds=self.creds)
        self._dc = conversation.DialogflowConversation(
            creds_path=creds_path, agent_id=agent_id
        )
        self._dffx = dataframe_functions.DataframeFunctions(creds=self.creds)

    def _build_sheets_client(self):
        client = gspread.authorize(self.creds)

        return client

    def _calculate_stats(self, df: pd.DataFrame):
        """Calculate all the stats needed for the summary report."""
        stats = Stats()
        stats.no_match_count = (
            df[df.detected_intent == "NO_MATCH"]
            .groupby("detected_intent")
            .size()
            .sum()
        )
        stats.no_match_rate = stats.no_match_count / df.shape[0]
        stats.pass_count = (
            df[df.detected_intent == df.expected_intent]
            .groupby("detected_intent")
            .size()
            .sum()
        )
        stats.pass_rate = stats.pass_count / df.shape[0]
        stats.test_agent = df.agent_display_name.unique()[0]
        stats.data_source = df.input_source.unique()[0]

        return stats

    def _write_report_summary_to_sheets(
        self, df: pd.DataFrame, sheet_name: str, sheet_tab: str
    ):
        """Writes the output report summary to Google Sheets."""

        gsheet = self._sheets_client.open(sheet_name)
        sheet = gsheet.worksheet(sheet_tab)

        df["test_run_timestamp"] = df.test_run_timestamp.astype("str")

        sheet.append_row(
            df.values.flatten().tolist(), value_input_option="USER_ENTERED"
        )

    def _append_test_results_to_sheets(
        self, results: pd.DataFrame, sheet_name: str, sheet_tab: str
    ):
        """Adds results to an existing Google Sheet collection."""

        gsheet = self._sheets_client.open(sheet_name)
        sheet = gsheet.worksheet(sheet_tab)

        # Fixes an error that sometimes happens when trying to write parameters
        # to the sheet because they are formatted as objects
        result_list = results.values.tolist()
        result_list = [list(map(str, row)) for row in result_list]

        sheet.append_rows(result_list, value_input_option="USER_ENTERED")

    def _write_test_results_to_sheets(
        self, results: pd.DataFrame, sheet_name: str, sheet_tab: str
    ):
        """Writes the output result details to Google Sheets."""

        gsheet = self._sheets_client.open(sheet_name)
        sheet = gsheet.worksheet(sheet_tab)

        sheet.clear()

        self._dffx.dataframe_to_sheets(sheet_name, sheet_tab, results)

    def _clean_dataframe(self, df):
        """Various Dataframe cleaning functions."""
        df.columns = df.columns.str.lower()
        df = df.replace("Start Page", "START_PAGE")
        df.rename(
                columns={
                    "source": "description",
                },
                inplace=True,
            )

        # Validate input schema
        try:
            df = df[INPUT_SCHEMA_COLUMNS]
        except KeyError as err:
            raise UserWarning("Ensure your input data contains the following "\
                              f"columns: {INPUT_SCHEMA_COLUMNS}") from err

        df["agent_display_name"] = self._a.get_agent(self.agent_id).display_name

        return df

    def process_input_csv(self, input_file_path: str):
        """Process the input data in CSV format."""
        df = pd.read_csv(input_file_path)
        df = df.fillna("")
        df = self._clean_dataframe(df)
        df["input_source"] = input_file_path

        return df

    def process_input_google_sheet(self, gsheet_name: str, gsheet_tab: str):
        """Process the input data in Google Sheets format."""
        df = self._dffx.sheets_to_dataframe(gsheet_name, gsheet_tab)
        df = self._clean_dataframe(df)
        df["input_source"] = gsheet_tab

        return df

    def run_evals(self, df: pd.DataFrame, chunk_size: int = 300,
                  rate_limit: float = 10.0,
                  eval_run_display_name: str = "Evals"):
        """Run the full Eval dataset."""
        logsx = "-" * 10

        logging.info(f"{logsx} STARTING {eval_run_display_name} {logsx}")
        results = self._dc.run_intent_detection(
            test_set=df, chunk_size=chunk_size, rate_limit=rate_limit
        )

        # Reorder Columns
        results = results.reindex(columns=OUTPUT_SCHEMA_COLUMNS)

        # When a NO_MATCH occurs, the detected_intent field will be blank
        # this replaces with NO_MATCH string, which will allow for easier stats
        # calculation downstream
        results.detected_intent.replace({"": "NO_MATCH"}, inplace=True)

        logging.info(f"{logsx} {eval_run_display_name} COMPLETE {logsx}")

        return results

    def generate_report(self, df: pd.DataFrame,
                        report_timestamp: datetime.datetime
                        ):
        """Generates a summary stats report for most recent NLU Eval tests."""
        # Calc fields
        stats = self._calculate_stats(df)

        # Generate Dataframe format
        df_report = pd.DataFrame(
            columns=SUMMARY_SCHEMA_COLUMNS,
            data=[
                [
                    report_timestamp,
                    df.shape[0],
                    stats.pass_count,
                    stats.pass_rate,
                    stats.no_match_count,
                    stats.no_match_rate,
                    stats.test_agent,
                    stats.data_source,
                ]
            ],
        )

        return df_report

    def write_summary_to_file(self, df: pd.DataFrame, output_file: str):
        """Write summary output to a local CSV file."""
        report_timestamp = datetime.datetime.now()
        df_report = self.generate_report(df, report_timestamp)
        df_report.to_csv(output_file, index=False)

    def write_results_to_file(self, df: pd.DataFrame, output_file: str):
        df.to_csv(output_file, index=False)

    def write_results_to_sheets(self, df: pd.DataFrame, google_sheet_name: str,
                                full_output_tab: str,
                                summary_tab: str,
                                append=False):
        """Write summary and detailed output to Google Sheets."""
        report_timestamp = datetime.datetime.now()
        df_report = self.generate_report(df, report_timestamp)

        self._write_report_summary_to_sheets(
            df_report, google_sheet_name, summary_tab
        )

        if append:
            self._append_test_results_to_sheets(
                df, google_sheet_name, full_output_tab
            )

        else:
            self._write_test_results_to_sheets(
                df, google_sheet_name, full_output_tab
            )



"""Utils for Cloud Storage and local file manipulation."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import zipfile
from google.cloud import storage
from google.oauth2 import service_account


class GcsUtils:
    """Utils for Cloud Storage and local file manipulation."""

    def __init__(self, creds_path: str = None, project_id: str = None):
        if creds_path and project_id:
            self.creds = service_account.Credentials.from_service_account_file(
                creds_path
            )
            self.gcs_client = storage.Client(
                credentials=self.creds, project=project_id
            )

        else:
            self.gcs_client = storage.Client()

    @staticmethod
    def unzip(agent_zip_file_path: str, extract_path: str):
        """Unzip file locally."""
        with zipfile.ZipFile(agent_zip_file_path, "r") as zip_ref:
            zip_ref.extractall(extract_path)

    @staticmethod
    def check_for_gcs_file(file_path: str) -> bool:
        """Validates GCS path vs. local path."""
        is_gcs_file = False

        file_prefix = file_path.split("/")[0]
        if file_prefix == "gs:":
            is_gcs_file = True

        return is_gcs_file

    def download_gcs(self, gcs_path: str, local_path: str = None):
        """Downloads the specified GCS file to local machine."""
        path = gcs_path.split("//")[1]
        bucket = path.split("/", 1)[0]
        gcs_object = path.split("/", 1)[1]
        file_name = gcs_object.split("/")[-1]
        bucket = self.gcs_client.bucket(bucket)
        blob = storage.Blob(gcs_object, bucket)

        if local_path:
            file_name = local_path + "/" + file_name

        blob.download_to_filename(file_name)

        return file_name



"""A set of builder methods to create CX proto resource objects"""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Dict

from google.cloud.dialogflowcx_v3beta1.types import Fulfillment
from google.cloud.dialogflowcx_v3beta1.types import ResponseMessage
from dfcx_scrapi.builders.builders_common import BuildersCommon
from dfcx_scrapi.builders.response_messages import ResponseMessageBuilder

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class FulfillmentBuilder(BuildersCommon):
    """Base Class for CX Fulfillment builder."""
    # TODO: ConditionalCases: def add_conditional_case(self) -> Fulfillment:

    _proto_type = Fulfillment
    _proto_type_str = "Fulfillment"


    def __str__(self) -> str:
        """String representation of the proto_obj."""
        try:
            self._check_proto_obj_attr_exist()
        except ValueError:
            return ""

        basic_info_str = self._show_basic_info()
        resp_msgs_str = self._show_response_messages()
        params_str = self._show_parameters()

        return (
            f"Fulfillment Basic Information:\n{'-'*20}\n{basic_info_str}"
            f"\n\n\nFulfillment ResponseMessages:\n{'-'*20}\n{resp_msgs_str}"
            f"\n\n\nFulfillment Parameters:\n{'-'*20}\n{params_str}"
        )


    def _show_basic_info(self) -> str:
        """String representation for the basic information of proto_obj."""
        self._check_proto_obj_attr_exist()

        partial_resp = self.proto_obj.return_partial_responses
        return (
            f"webhook: {self.proto_obj.webhook}"
            f"\ntag: {self.proto_obj.tag}"
            f"\nreturn_partial_responses: {partial_resp}"
        )


    def _show_parameters(self) -> str:
        """String representation for the parameters presets of proto_obj."""
        self._check_proto_obj_attr_exist()

        return "\n".join([
            f"{param.parameter}: {param.value if param.value else 'null'}"
            for param in self.proto_obj.set_parameter_actions
        ])


    def _show_response_messages(self) -> str:
        """String representation of response messages in proto_obj."""
        self._check_proto_obj_attr_exist()

        return "\n".join([
            f"ResponseMessage {i+1}:\n{str(ResponseMessageBuilder(msg))}"
            for i, msg in enumerate(self.proto_obj.messages)
        ])


    def show_fulfillment(self, mode: str = "whole"):
        """Show the proto_obj information.
        Args:
          mode (str):
            Specifies what part of the fulfillment to show.
            Options:
              ['basic', 'parameters',
                'messages' or 'response messages', 'whole']
        """
        self._check_proto_obj_attr_exist()

        if mode == "basic":
            print(self._show_basic_info())
        elif mode == "parameters":
            print(self._show_parameters())
        elif mode in ["messages", "response messages"]:
            print(self._show_response_messages())
        elif mode == "whole":
            print(self)
        else:
            raise ValueError(
                "mode should be in"
                "['basic', 'parameters',"
                " 'messages' or 'response messages', 'whole']"
            )


    def create_new_proto_obj(
        self,
        webhook: str = None,
        tag: str = None,
        return_partial_responses: bool = False,
        overwrite: bool = False
    ) -> Fulfillment:
        """Create a new Fulfillment.

        Args:
          webhook (str):
            The webhook to call. Format:
            ``projects/<Project ID>/locations/<Location ID>/agents
              /<Agent ID>/webhooks/<Webhook ID>``.
          tag (str):
            The tag is typically used by
            the webhook service to identify which fulfillment is being
            called, but it could be used for other purposes. This field
            is required if ``webhook`` is specified.
          return_partial_responses (bool):
            Whether Dialogflow should return currently
            queued fulfillment response messages in
            streaming APIs. If a webhook is specified, it
            happens before Dialogflow invokes webhook.
          overwrite (bool)
            Overwrite the new proto_obj if proto_obj already
            contains a Fulfillment.

        Returns:
            A Fulfillment object stored in proto_obj.
        """
        # Types error checking
        if (return_partial_responses and
            not isinstance(return_partial_responses, bool)):
            raise ValueError(
                "return_partial_responses should be bool."
            )
        if ((webhook and not isinstance(webhook, str)) or
            (tag and not isinstance(tag, str))):
            raise ValueError(
                "webhook and tag should be string."
            )
        # webhook with tag presence check
        if webhook and not tag:
            raise ValueError(
                "tag is required when webhook is specified."
            )
        # `overwrite` parameter error checking
        if self.proto_obj and not overwrite:
            raise UserWarning(
                "proto_obj already contains a Fulfillment."
                " If you wish to overwrite it, pass overwrite as True."
            )
        # Create the fulfillment
        if overwrite or not self.proto_obj:
            self.proto_obj = Fulfillment(
                webhook=webhook,
                return_partial_responses=return_partial_responses,
                tag=tag
            )

        return self.proto_obj


    def add_response_message(
        self,
        response_message: ResponseMessage
    ) -> Fulfillment:
        """Add a rich message response to present to the user.

        Args:
          response_message (ResponseMessage):
            The ResponseMessage to add to the Fulfillment.
            Refer to `builders.response_message.ResponseMessageBuilder`
              to build one.

        Returns:
          A Fulfillment object stored in proto_obj
        """
        self._check_proto_obj_attr_exist()

        if not isinstance(response_message, ResponseMessage):
            raise ValueError(
                "`response_message` type should be ResponseMessage."
            )

        self.proto_obj.messages.append(response_message)

        return self.proto_obj


    def add_parameter_presets(
        self,
        parameter_map: Dict[str, str]
    ) -> Fulfillment:
        """Set parameter values before executing the webhook.

        Args:
          parameter_map (Dict[str, str]):
            A dictionary that represents parameters as keys
            and the parameter values as it's values.
            A `None` value clears the parameter.

        Returns:
          A Fulfillment object stored in proto_obj
        """
        self._check_proto_obj_attr_exist()

        # Type error checking
        if isinstance(parameter_map, dict):
            if not all((
                isinstance(key, str) and isinstance(val, str)
                for key, val in parameter_map.items()
            )):
                raise ValueError(
                    "Only strings are allowed as"
                    " dictionary keys and values in parameter_map."
                )
            for param, val in parameter_map.items():
                self.proto_obj.set_parameter_actions.append(
                    Fulfillment.SetParameterAction(parameter=param, value=val)
                )

            return self.proto_obj
        else:
            raise ValueError(
                "parameter_map should be a dictionary."
            )


    def remove_parameter_presets(
        self,
        parameter_map: Dict[str, str]
    ) -> Fulfillment:
        """Remove parameter values from the fulfillment.

        Args:
          parameter_map (Dict[str, str]):
            A dictionary that represents parameters as keys
            and the parameter values as it's values.
            A `None` value clears the parameter.

        Returns:
          A Fulfillment object stored in proto_obj
        """
        self._check_proto_obj_attr_exist()

        # Type error checking
        if isinstance(parameter_map, dict):
            if not all((
                isinstance(key, str) and isinstance(val, str)
                for key, val in parameter_map.items()
            )):
                raise ValueError(
                    "Only strings are allowed as"
                    " dictionary keys and values in parameter_map."
                )

            new_params = []
            for param in self.proto_obj.set_parameter_actions:
                if (
                    param.parameter in parameter_map and
                    param.value == parameter_map[param.parameter]
                ):
                    continue

                new_params.append(param)

            self.proto_obj.set_parameter_actions = new_params
            return self.proto_obj
        else:
            raise ValueError(
                "parameter_map should be a dictionary."
            )


    def has_webhook(self) -> bool:
        """Check whether the Fulfillment in proto_obj uses a Webhook.

        Returns:
          True if proto_obj uses a Webhook and False otherwise
        """
        try:
            self._check_proto_obj_attr_exist()
        except ValueError:
            return False

        return bool(self.proto_obj.webhook)



"""Collection of Type Classes used for offline processing."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Any, Tuple
from dataclasses import dataclass, field

from dfcx_scrapi.agent_extract import graph as graph_class

@dataclass
class AgentMetadata:
    """Used to track the current Agent Metadata attrinbutes."""

    default_language_code: str = None
    dtmf_settings: bool = False
    logging_enabled: bool = False
    speech_adaptation: bool = False


@dataclass
class Flow:
    """Used to track current Flow Attributes."""
    agent_id: str = None
    all_pages: set = field(default_factory=set)
    active_pages: set = field(default_factory=set)
    data: Dict[str, Any] = field(default_factory=dict)
    dangling_pages: set = field(default_factory=set)
    dir_path: str = None  # Full Directory Path for this Flow
    display_name: str = None  # Flow Display Name (removed special chars)
    file_name: str = None  # Original Name of Flow (includes special chars)
    graph: graph_class.Graph = None
    resource_id: str = None
    resource_type: str = "flow"
    start_page_file: str = None  # File Path Location of START_PAGE
    unreachable_pages: set = field(default_factory=set)
    unused_pages: set = field(default_factory=set)

@dataclass
class Page:
    """Used to track current Page Attributes."""

    agent_id: str = None
    data: Dict[str, Any] = None
    display_name: str = None
    entry: Dict[str, Any] = None
    events: List[object] = None
    flow: Flow = None
    form: Dict[str, Any] = None
    has_webhook: bool = False
    has_webhook_event_handler: bool = False
    page_file: str = None
    resource_id: str = None
    resource_type: str = "page"
    routes: List[object] = None
    route_groups: List[str] = None

@dataclass
class FormParameter:
    """Tracks Form Paramter attributes within a Page."""

    advanced_settings: str = None
    agent_id: str = None
    data: Dict[str, Any] = None
    display_name: str = None
    dtmf_settings: str = None
    entity_type: str = None
    fill_behavior: Dict[str, Any] = None
    init_fulfillment: Dict[str, Any] = None
    page: Page = None
    reprompt_handlers: Dict[str, Any] = None
    required: bool = True


@dataclass
class RouteGroup:
    """Used to track current RouteGroup Attributes."""

    agent_id: str = None
    data: Dict[str, Any] = None
    display_name: str = None
    flow: Flow = None
    resource_id: str = None
    resource_type: str = "route_group"
    rg_file: str = None
    routes: List[object] = None

@dataclass
class Fulfillment:
    """Used to track current Fulfillment Attributes."""

    agent_id: str = None
    data: Dict[str, Any] = None
    display_name: str = None  # Inherit from Page easy logging
    fulfillment_type: str = None  # transition_route | event
    page: Page = None
    parameter: str = None # Used for Reprompt Event Handlers
    target_flow: str = None
    target_page: str = None
    text: str = None
    trigger: str = None
    resource_type: str = "fulfillment"

@dataclass
class Intent:
    """Used to track current Intent Attributes."""

    agent_id: str = None
    data: Dict[str, Any] = None
    description: str = None
    display_name: str = None
    dir_path: str = None
    labels: Dict[str, str] = None
    metadata_file: str = None
    parameters: List[Dict[str, str]] = field(default_factory=list)
    resource_id: str = None
    resource_type: str = "intent"
    training_phrases: Dict[str, Any] = field(default_factory=dict)

@dataclass
class EntityType:
    """Used to track current Flow Attributes."""

    agent_id: str = None
    auto_expansion: str = None
    data: Dict[str, Any] = None
    dir_path: str = None  # Full Directory Path for this Entity Type
    display_name: str = None  # Entity Type Display Name
    entities: Dict[str, Any] = field(default_factory=dict)  # Map
    excluded_phrases: Dict[str, Any] = field(default_factory=dict)  # Map
    fuzzy_extraction: bool = False
    kind: str = None  # The kind of Entity Type represented
    resource_id: str = None
    resource_type: str = "entity_type"

@dataclass
class TestCase:
    """Used to track current Test Case Attributes."""

    associated_intent_data: Dict[str, Any] = None
    agent_id: str = None
    agent_path: str = None
    conversation_turns: List[Any] = None
    data: Dict[str, Any] = None
    dir_path: str = None
    display_name: str = None
    has_invalid_intent: bool = False
    intent_data: List[str] = None
    qualified: bool = False
    resource_id: str = None
    resource_type: str = "test_case"
    tags: List[str] = None
    test_config: Dict[str, Any] = None

@dataclass
class Webhook:
    """Used to track current Webhook attributes."""

    agent_id: str = None
    agent_path: str = None
    data: Dict[str, Any] = None
    dir_path: str = None
    display_name: str = None
    resource_id: str = None
    resource_type: str = "webhook"
    service_type: str = None
    timeout: int = 0

@dataclass
class AgentData:
    """Used to track agent data for each section processed."""
    active_intents: Dict[str, List[Tuple[str, str]]] = field(
        default_factory=dict)
    active_pages: Dict[str, set] = field(default_factory=dict)
    agent_id: str = None
    entity_types: List[Dict[str, Any]] = field(default_factory=list)
    entity_types_map: Dict[str, Any] = field(default_factory=dict)
    flow_page_map: Dict[str, Any] = field(default_factory=dict)
    flows: List[Dict[str, Any]] = field(default_factory=list)
    flows_map: Dict[str, Any] = field(default_factory=dict)
    graph: graph_class.Graph = None
    intents: List[Dict[str, Any]] = field(default_factory=list)
    intents_map: Dict[str, Any] = field(default_factory=dict)
    lang_code: str = "en"
    pages: Dict[str, List[Dict[str, Any]]] = field(default_factory=dict)
    route_groups: Dict[str, List[Dict[str, Any]]] = field(default_factory=dict)
    route_groups_map: Dict[str, Any] = field(default_factory=dict)
    test_cases: List[Dict[str, Any]] = field(default_factory=list)
    unreachable_pages: Dict[str, set] = field(default_factory=dict)
    unused_pages: Dict[str, set] = field(default_factory=dict)
    webhooks: List[Dict[str, Any]] = field(default_factory=list)
    webhooks_map: Dict[str, Any] = field(default_factory=dict)

    total_flows: int = 0
    total_pages: int = 0
    total_intents: int = 0
    total_training_phrases: int = 0
    total_entity_types: int = 0
    total_route_groups: int = 0
    total_test_cases: int = 0
    total_webhooks: int = 0



"""Methods for creating CX object types"""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# DEPRECATION WARNING!
# THIS CLASS WILL BE DEPRECATED IN A FUTURE VERSION OF SCRAPI POST v1.5.1
# PLEASE NOTE THAT THE /builders CLASSES WILL TAKE PRECEDENCE OVER THIS CLASS

import logging
from google.cloud.dialogflowcx_v3beta1 import types

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

SCOPES = [
    "https://www.googleapis.com/auth/cloud-platform",
    "https://www.googleapis.com/auth/dialogflow",
]


class MakerUtil:
    """Util class to create CX objects like transition routes"""
    @classmethod
    def make_generic(cls, obj, obj_type, default, conditionals=None):
        if conditionals is None:
            conditionals = {}

        if isinstance(obj, obj_type):
            return obj

        elif isinstance(obj, dict):
            obj_ins = obj_type()
            for key, value in obj.items():
                if key in conditionals.keys():
                    func = conditionals[key]
                    out = func(value)
                    setattr(obj_ins, key, out)
                else:
                    print(value)
                    setattr(obj_ins, key, value)
            return obj_ins

        elif isinstance(obj, str):
            dic = {
                "unspecified": 0,
                "map": 1,
                "list": 2,
                "regexp": 3,
                "default": 1,
            }
            t = dic.get(obj.lower())
            if t:
                return obj_type(t)
            else:
                return default
        else:
            return default

    @classmethod
    def make_seq(cls, obj, obj_type, default, conditionals=None):
        if conditionals is None:
            conditionals = {}
        assert isinstance(obj, list)
        l = []
        for x in obj:
            l.append(cls.make_generic(
                x, obj_type, default, conditionals))
        return l

    @classmethod
    def make_transition_route(cls, obj=None, **kwargs):
        """Creates a single Transition Route object for Dialogflow CX.

        Transition routes are used to navigate a user from page to page, or
        page to flow in Dialogflow CX. Routes can be part of a Page object or
        they can also be associated with Route Groups. In either case, the
        structure of the Route is the same. This method allows the user to
        create a single Route object that can be used interchangeably with
        Pages or Route Groups as needed.

        Note: if no args are provided, a blank Route object will be created.

        Args:
          obj: (Optional) an existing Route object can be provided if the
            user wants to modify or duplicate the object.

        Keyword Args:
          intent, (str): The UUID of the Intent to route to
          condition, (str): The condition to evaluate on the route
          target_page, (str): The UUID of the target page to transition to
          target_flow, (str): The UUID of the target flow to transition to
          trigger_fulfillment, (obj): Requires an object in the format of type
            <google.cloud.dialogflowcx_v3beta1.types.fulfillment.Fulfillment>

        Returns:
          Route object of type
          <google.cloud.dialogflowcx_v3beta1.types.page.TransitionRoute>
        """

        if obj:
            route = obj

            # make sure the route name is cleared if this is a copy of
            # another existing route object
            route.name = ""

        else:
            route = types.page.TransitionRoute()

        # Set route attributes to args
        for key, value in kwargs.items():
            if key == "trigger_fulfillment":
                tf = cls.make_trigger_fulfillment(value)
                setattr(route, key, tf)
            else:
                setattr(route, key, value)

        return route

    @classmethod
    def make_trigger_fulfillment(
        cls, messages=None, webhook_id=None, webhook_tag=None
    ):
        """Creates a single Fulfillment object for Dialogflow CX.

        Fulfillments are used as part of Transition Routes to add Dialogue
        messages back to the user, trigger webhooks, set parameter presets,
        and enable IVR options where applicable.

        Note: if no args are provided, a blank Fulfillment object will be
        returned.

        Args:
          messages: Optional list of Dialogue messages to send
          webhook_id, (str): (Optional)
            The UUID of the Dialogflow CX webhook to trigger
            when the Fulfillment is triggered by the conversation.
          webhook_tag, (str): (Required if webhook_id is provided)
            User defined tag associated with

        Returns:
          Fulfillment object of type
          <google.cloud.dialogflowcx_v3beta1.types.fulfillment.Fulfillment>
        """
        fulfillment = types.fulfillment.Fulfillment()

        if messages:
            response_message = types.response_message.ResponseMessage()
            message_text = response_message.Text()

            message_text.text = messages
            response_message.text = message_text
            fulfillment.messages = [response_message]

        if webhook_id:
            fulfillment.webhook = webhook_id

            if not webhook_tag:
                logging.info(
                    "webhook_tag is required when specifying webhook_id")
                return None

            else:
                fulfillment.tag = webhook_tag

        # print(fulfillment)
        return fulfillment

    @classmethod
    def set_entity_type_attr(cls, entity_type, kwargs):
        for key, value in kwargs.items():
            if key == "kind":
                kind = types.entity_type.EntityType.Kind
                obj = cls.make_generic(value, kind, kind(0))
                setattr(entity_type, key, obj)
            # For the auto expansion mode case create helper object to set at
            # entity_type attribute
            elif key == "auto_expansion_mode":
                aem = types.entity_type.EntityType.AutoExpansionMode
                obj = cls.make_generic(value, aem, aem(1))
                setattr(entity_type, key, obj)

            # For the entities case iterate over dictionary and assign key value
            # pairs to entity type elements of entities list
            elif key == "entities":
                entity = types.entity_type.EntityType.Entity
                obj = cls.make_seq(value, entity, entity())
                setattr(entity_type, key, obj)

            # For the excluded phrases case assign value to the excluded phrase
            # object then set as the entity_type attribute
            elif key == "excluded_phrases":
                ep = types.entity_type.EntityType.ExcludedPhrase
                obj = cls.make_seq(value, ep, ep())
                setattr(entity_type, key, obj)

            else:
                setattr(entity_type, key, value)



"""Example of building Dialogflow CX Agents using SCRAPI library."""
#!/usr/bin/env python
# coding: utf-8

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import pandas as pd
from dfcx_scrapi.core.agents import Agents
from dfcx_scrapi.tools.dataframe_functions import DataframeFunctions

# # Method 1 - Creating an Agent from Simple Text Inputs

# ## Create Your Agent
# Creating an agent requires a minimum of 2 pieces of information:
# - `project_id`, which is your GCP Project ID
# - `display_name`, (i.e. 'My Cool Agent!')
# - `gcp_region`, (Optional) This defaults to `global` region, but you can
# 	provide any GCP region that is currently available for Dialogflow CX.


def build_agent(creds_path, project_id, gcp_region, display_name):
    """Build out CX Agent."""

    # First we will instantiate our Agent object
    agent_instance = Agents(creds_path=creds_path)

    # Next, we will set some variables for our agent creation or retrieval args

    # Then we will call the `create_agent` and capture the result in a var call
	# `my_agent`
    my_agent = agent_instance.create_agent(project_id, display_name, gcp_region)

    # Option 2: If agent already exists
    # The agent ID must be entered as a string
    # "projects/<PROJECT_ID>/locations/<GCP_REGION>/agents/<AGENT_ID>"
    # my_agent = a.get_agent(
	#   "projects/<PROJECT_ID>/locations/<GCP_REGION>/agents/<AGENT_ID>")

    # ## Create Your First Intent
    # For this demo agent, we'll build a basic intent from list of Training
	# Phrases (TPs)

    # To simplify the Intent creation, we'll utilize the `DataframeFunctions`
    # class from the `tools` portion of the SCRAPI library.
    # This will allow us to build our intent into a simple Pandas DataFrame,
    # and then push this DataFrame directly into our bot that we just created.

    # A common method of building Intents and Training Phrases for Dialogflow
    # CX agents is to use Google Sheets or CSVs to store the Intent/TP data.
    # For this demo, we are working with annotated training phrases. To do so,
    # we've included
    # * `intent_sample_with_parts.csv` which contains all the training phrases
    # 	parts. It will be pulled into a dataframe
    # * `intent_sample_params`, that defines the parameters used for each
    # 	intent. It will also be pulled into a dataframe

    # First, we will instantiate our DataframeFunctions (dffx) object
    dffx = DataframeFunctions(creds_path=creds_path)

    # Next, we will read in our sample CSV with Intent training phrases data as
    # well as the entities used into 2 distinct Pandas DataFrames
    df = pd.read_csv("../../data/intent_sample_with_parts.csv")  # pylint: disable=C0103
    params_df = pd.read_csv("../../data/intent_sample_params.csv")

    # Finally, we will use `dffx` to push our Intents to our Agent
    # If intents do not exist, use bulk_create_intent_from_dataframe
    _ = dffx.bulk_create_intent_from_dataframe(
        my_agent.name, df, params_df, update_flag=True, mode="advanced"
    )


# If intent already exist, use bulk_update_intents_from_dataframe
# my_intents = dffx.bulk_update_intents_from_dataframe(
# my_agent.name, df, params_df, update_flag=True, mode="advanced")


# # Bot Building 101 End
# ## And there you have it!
# We've created a simple Dialogflow CX agent using only Python in a Jupyter
# notebook.
# You can see how this could be easily scaled up using .py files, git repos,
# and other scripts to speed up the bot building process.


# The agent ID must be entered as a string
# 	"projects/<PROJECT_ID>/locations/<GCP_REGION>/agents/<AGENT_ID>"
def update_agent(creds_path, agent_id):
    """Updating the agent."""
    # First we will instantiate our Agent object and retrieve the agent using
    # its id
    agent_instance = Agents(creds_path=creds_path)
    my_agent = agent_instance.get_agent(
        f"projects/<PROJECT_ID>/locations/<GCP_REGION>/agents/{agent_id}"
    )

    # ## Create Your First Intent
    # For this demo agent, we'll build a basic intent from list of Training
    # Phrases (TPs)

    # To simplify the Intent creation, we'll utilize the `DataframeFunctions`
    # class from the `tools` portion of the SCRAPI library. This will allow
    # us to build our intent into a simple Pandas DataFrame, and then push
    # this DataFrame directly into our bot that we just created.

    # A common method of building Intents and Training Phrases for Dialogflow
    # CX agents is to use Google Sheets or CSVs to store the Intent/TP data.
    # For this demo, we are working with annotated training phrases. To do so,
    # we've included:
    #   * `intent_sample_with_parts.csv` which contains all the training
    #     phrases parts. It will be pulled into a dataframe
    #   * `intent_sample_params`, that defines the parameters used for each
    #     intent. It will also be pulled into a dataframe

    # First, we will instantiate our DataframeFunctions (dffx) object
    dffx = DataframeFunctions(creds_path=creds_path)

    # Next, we will read in our sample CSV with Intent training phrases data as
    #  well as the entities used into 2 distinct Pandas DataFrames
    df = pd.read_csv("../../data/intent_sample_with_parts.csv")  # pylint: disable=C0103
    params_df = pd.read_csv("../../data/intent_sample_params.csv")

    # bulk update the existing intents
    _ = dffx.bulk_update_intents_from_dataframe(
        my_agent.name, df, params_df, update_flag=True, mode="advanced"
    )


# # Bot Building 102 End
# ## And there you have it!
# We've created a simple Dialogflow CX agent using only Python in a Jupyter
# notebook. You can see how this could be easily scaled up using .py files,
# git repos, and other scripts to speed up the bot building process.

if __name__ == "__main__":
    CREDS_PATH = str(sys.argv[1])
    PROJECT_ID = str(sys.argv[2])
    REGION = str(sys.argv[3])
    DISPLAY_NAME = str(sys.argv[4])

    build_agent(CREDS_PATH, PROJECT_ID, REGION, DISPLAY_NAME)



"""Test Cases Resource functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
import logging
from typing import Dict, List

from google.cloud.dialogflowcx_v3beta1 import services
from google.cloud.dialogflowcx_v3beta1 import types
from google.protobuf import field_mask_pb2

from dfcx_scrapi.core import scrapi_base
from dfcx_scrapi.core import flows
from dfcx_scrapi.core import pages

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class TestCases(scrapi_base.ScrapiBase):
    """Core Class for CX Test Cases."""

    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        scope=False,
        agent_id: str = None,
        test_case_id: str = None,
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

        if agent_id:
            self.agent_id = agent_id
            self.client_options = self._set_region(self.agent_id)

        if test_case_id:
            self.test_case_id = test_case_id
            self.client_options = self._set_region(self.test_case_id)

    def _convert_test_result_to_string(self, test_case: types.TestCase) -> str:
        """Converts the Enum result to a string."""
        if test_case.last_test_result.test_result == 0:
            return "TEST_RESULT_UNSPECIFIED"
        elif test_case.last_test_result.test_result == 1:
            return "PASSED"
        elif test_case.last_test_result.test_result == 2:
            return "FAILED"
        else:
            return ""

    def _convert_test_result_to_bool(self, test_case: types.TestCase) -> bool:
        """Converts the String result to a boolean."""
        test_result = self._convert_test_result_to_string(test_case)

        if test_result == "PASSED":
            return True
        elif test_result == "FAILED":
            return False
        else:
            return None

    def _get_flow_id_from_test_config(
            self, test_case: types.TestCase) -> str:
        """Attempt to get the Flow ID from the Test Case Test Config."""
        if "flow" in test_case.test_config:
            return test_case.test_config.flow
        elif "page" in test_case.test_config:
            return "/".join(test_case.test_config.page.split("/")[:8])
        else:
            agent_id = "/".join(test_case.name.split("/")[:6])
            return f"{agent_id}/flows/00000000-0000-0000-0000-000000000000"

    def _get_page_id_from_test_config(
            self, test_case: types.TestCase, flow_id: str) -> str:
        """Attempt to get the Page ID from the Test Case Test Config."""
        if "page" in test_case.test_config:
            return test_case.test_config.page
        else:
            return f"{flow_id}/pages/START_PAGE"

    def _get_page_display_name(
            self, flow_id: str, page_id: str,
            pages_map: Dict[str, Dict[str, str]]) -> str:
        """Get the Page Display Name from the Pages Map based on the Page ID."""
        page_map = pages_map.get(flow_id, None)
        page = "START_PAGE"
        if page_map:
            page = page_map.get(page_id, None)

        return page

    def _process_test_case(self, test_case, flows_map, pages_map):
        """Takes a response from list_test_cases and returns a single row
        dataframe of the test case result.

        Args:
          test_case: The test case response
          flows_map: A dictionary mapping flow IDs to flow display names
          pages_map: A dictionary with keys as flow IDs and values as
            dictionaries mapping page IDs to page display names for that flow

        Returns: A dataframe with columns:
          display_name, id, short_id, tags, creation_time, start_flow,
          start_page, test_result, passed, test_time
        """
        flow_id = self._get_flow_id_from_test_config(test_case)
        page_id = self._get_page_id_from_test_config(test_case, flow_id)
        page = self._get_page_display_name(flow_id, page_id, pages_map)
        test_result = self._convert_test_result_to_bool(test_case)

        return pd.DataFrame(
            {
                "display_name": [test_case.display_name],
                "id": [test_case.name],
                "short_id": [test_case.name.split("/")[-1]],
                "tags": [",".join(test_case.tags)],
                "creation_time": [test_case.creation_time],
                "start_flow": [flows_map.get(flow_id, None)],
                "start_page": [page],
                # "test_result": [test_result],
                "passed": [test_result],
                "test_time": [test_case.last_test_result.test_time]
            }
        )

    def _retest_cases(
            self, test_case_df: pd.DataFrame, retest_ids: List[str]
            ) -> pd.DataFrame:
        print("To retest:", len(retest_ids))
        response = self.batch_run_test_cases(retest_ids, self.agent_id)
        for result in response.results:
            # Results may not be in the same order as they went in
            # Process the name a bit to remove the /results/id part
            tc_id_full = "/".join(result.name.split("/")[:-2])
            tc_id = tc_id_full.rsplit("/", maxsplit=1)[-1]

            # Update dataframe where id = tc_id_full
            # row = test_case_df.loc[test_case_df['id']==tc_id_full]
            test_case_df.loc[
                test_case_df["id"] == tc_id_full, "short_id"
            ] = tc_id
            # test_case_df.loc[
            #     test_case_df["id"] == tc_id_full, "test_result"
            # ] = str(result.test_result)
            test_case_df.loc[
                test_case_df["id"] == tc_id_full, "test_time"
            ] = result.test_time
            test_case_df.loc[test_case_df["id"] == tc_id_full,"passed"] = (
                str(result.test_result) == "TestResult.PASSED"
            )

        return test_case_df

    @scrapi_base.api_call_counter_decorator
    def list_test_cases(
        self, agent_id: str = None, include_conversation_turns: bool = False
    ):
        """List test cases from an agent.

        Args:
          agent_id: The agent to list all test cases for.
            `projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>`
          include_conversation_turns: Either to include the conversation turns
            in the test cases or not. Default is False
            which shows only the basic metadata about the test cases.

        Returns:
          List of test cases from an agent.
        """

        if not agent_id:
            agent_id = self.agent_id

        if include_conversation_turns:
            test_case_view = types.ListTestCasesRequest.TestCaseView.FULL
        else:
            test_case_view = types.ListTestCasesRequest.TestCaseView.BASIC

        request = types.test_case.ListTestCasesRequest(
            parent=agent_id, view=test_case_view
        )

        client_options = self._set_region(agent_id)

        client = services.test_cases.TestCasesClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.list_test_cases(request)

        test_cases = []
        for page in response.pages:
            for test_case in page.test_cases:
                test_cases.append(test_case)

        return test_cases

    @scrapi_base.api_call_counter_decorator
    def export_test_cases(
        self,
        gcs_uri: str,
        agent_id: str = None,
        data_format: str = None,
        data_filter: str = None,
    ):
        """Export test cases from an agent to cloud storage

        Args:
          gcs_uri: The GCS URI to export the test cases to. The format of this
            URI must be `gs://<bucket-name>/<object-name>`. If unspecified,
            the serialized test cases are returned inline.
          agent_id: The agent to export test cases from.
            `projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>`
          data_format: The data format of the exported test cases. If not
            specified, `BLOB` is assumed.
          data_filter: The filter expression used to filter exported test
            cases, see `API Filtering <https://aip.dev/160>`__. The expression
            is case insensitive and supports the following syntax:
              name = [OR name = ] ...
              For example:
                -  "name = t1 OR name = t2" matches the test case with the
                exact resource name "t1" or "t2".

        Returns:
          Long running operation for export
        """

        if not agent_id:
            agent_id = self.agent_id

        request = types.test_case.ExportTestCasesRequest()
        request.parent = agent_id
        request.gcs_uri = gcs_uri
        request.data_format = data_format
        request.filter = data_filter

        client_options = self._set_region(agent_id)
        client = services.test_cases.TestCasesClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.export_test_cases(request)

        return response

    @scrapi_base.api_call_counter_decorator
    def create_test_case(self, test_case: types.TestCase, agent_id: str = None):
        """Create a new Test Case in the specified CX Agent.

        Args:
          test_case: The test case to create.
          agent_id: The agent to create the test case for. Format:
              `projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>`

        Returns:
          The test case which was created
        """
        if not agent_id:
            agent_id = self.agent_id

        request = types.test_case.CreateTestCaseRequest()
        request.parent = agent_id
        request.test_case = test_case

        client_options = self._set_region(agent_id)
        client = services.test_cases.TestCasesClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.create_test_case(request)
        return response

    @scrapi_base.api_call_counter_decorator
    def get_test_case(self, test_case_id: str):
        """Get test case object from CX Agent.

        Args:
          test_case_id: The name of the test case. Format:
            `projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
              testCases/<TestCase ID>`

        Returns:
          The test case object
        """

        request = types.test_case.GetTestCaseRequest()
        request.name = test_case_id

        client_options = self._set_region(test_case_id)
        client = services.test_cases.TestCasesClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.get_test_case(request)
        return response

    @scrapi_base.api_call_counter_decorator
    def import_test_cases(self, gcs_uri: str, agent_id: str = None):
        """Import test cases from cloud storage.

        Args:
          gcs_uri: The GCS URI to import test cases from. The format of this
            URI must be `gs://<bucket-name>/<object-name>`
          agent_id: The agent to import test cases to. Format:
            `projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>`

        Returns:
          Long running operation for importing test cases.
        """

        if not agent_id:
            agent_id = self.agent_id

        request = types.test_case.ImportTestCasesRequest()
        request.parent = agent_id
        request.gcs_uri = gcs_uri

        client_options = self._set_region(agent_id)
        client = services.test_cases.TestCasesClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.import_test_cases(request)
        result = response.result()
        return result

    @scrapi_base.api_call_counter_decorator
    def batch_delete_test_cases(
        self,
        test_case_ids: List[str],
        agent_id: str = None):
        """Delete a set of test cases from an agent.

        Args:
          test_case_id: List of test case names in the following format:
            `projects/<Project ID>/locations/ <Location ID>/agents/<AgentID>/
              testCases/<TestCase ID>`
          agent_id: The agent to delete test cases from. Format:
            `projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>`.

        Returns:
          None
        """

        if not agent_id:
            agent_id = self.agent_id

        request = types.test_case.BatchDeleteTestCasesRequest()
        request.parent = agent_id
        request.names = test_case_ids

        client_options = self._set_region(agent_id)
        client = services.test_cases.TestCasesClient(
            credentials=self.creds, client_options=client_options
        )
        client.batch_delete_test_cases(request)

    @scrapi_base.api_call_counter_decorator
    def list_test_case_results(self, test_case_id: str):
        """List the results from a specific Test Case.

        Args:
          test_case_id: The test case to list results for. Format:
            `projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
              testCases/<TestCase ID>`
            NOTE: Specify a ``-`` as a wildcard for TestCase ID to list
            results across multiple test cases.

        Returns:
          List of test case results
        """

        request = types.test_case.ListTestCaseResultsRequest()
        request.parent = test_case_id

        client_options = self._set_region(test_case_id)
        client = services.test_cases.TestCasesClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.list_test_case_results(request)

        test_case_results = []
        for page in response.pages:
            for result in page.test_case_results:
                test_case_results.append(result)

        return test_case_results

    @scrapi_base.api_call_counter_decorator
    def batch_run_test_cases(
        self,
        test_cases: List[str],
        agent_id: str = None,
        environment: str = None):
        """Run a set of test cases to get their latest results.

        Args:
          test_cases: List of Test Case IDs in the following format:
            `projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
              testCases/<TestCase ID>`
          agent_id: The CX Agent ID to run the Test Cases on.
            `projects/<Project ID>/locations/<Location ID>/agents/<AgentID>`
          environment: If not set, draft environment is assumed. Format:
            `projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
            environments/<Environment ID>`

        Returns:
          Results for the set of run test cases.
        """

        if not agent_id:
            agent_id = self.agent_id

        request = types.test_case.BatchRunTestCasesRequest()
        request.parent = agent_id
        request.environment = environment
        request.test_cases = test_cases

        client_options = self._set_region(agent_id)
        client = services.test_cases.TestCasesClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.batch_run_test_cases(request)
        results = response.result()
        return results

    @scrapi_base.api_call_counter_decorator
    def update_test_case(
        self,
        test_case_id: str = None,
        obj: types.TestCase = None,
        **kwargs) -> types.TestCase:
        """Update Test Case attributes for a specified Test Case.

        Args:
          test_case_id: The Test Case ID to update.
          obj: The Test Case obj of types.TestCase to use for the update.

        Returns:
          The updated Test Case.
        """

        if obj:
            test_case = obj
            test_case.name = test_case_id
        else:
            if not test_case_id:
                test_case_id = self.test_case_id
            test_case = self.get_test_case(test_case_id)

        for key, value in kwargs.items():
            setattr(test_case, key, value)
        paths = kwargs.keys()
        mask = field_mask_pb2.FieldMask(paths=paths)

        request = types.test_case.UpdateTestCaseRequest()
        request.test_case = test_case
        request.update_mask = mask

        client_options = self._set_region(test_case_id)
        client = services.test_cases.TestCasesClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.update_test_case(request)
        return response

    @scrapi_base.api_call_counter_decorator
    def run_test_case(self, test_case_id: str, environment: str = None):
        """Run test case and get result for a specified test case.

        Args:
          test_case_id: Test Case ID in the following format:
            `projects/<Project ID>/locations/ <Location ID>/agents/<AgentID>/
              testCases/<TestCase ID>`
          environment: The CX Environment name. If not set, DRAFT environment
            is assumed. Format: `projects/<Project ID>/locations/<Location ID>/
              agents/<Agent ID>/environments/<Environment ID>`

        Returns:
          The test case result.
        """

        request = types.test_case.RunTestCaseRequest()
        request.name = test_case_id
        request.environment = environment

        client_options = self._set_region(test_case_id)
        client = services.test_cases.TestCasesClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.run_test_case(request)
        results = response.result()
        return results

    @scrapi_base.api_call_counter_decorator
    def get_test_case_result(self, test_case_result_id: str):
        """Get test case result for a specified run on a specified test case.

        Args:
          test_case_result_id: The Test Case Result ID to retrieve.
            projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
              testCases/<TestCase ID>/results/<TestCaseResult ID>

        Returns:
          The test case result.
        """
        request = types.test_case.GetTestCaseResultRequest()
        request.name = test_case_result_id

        client_options = self._set_region(test_case_result_id)
        client = services.test_cases.TestCasesClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.get_test_case_result(request)
        return response

    @scrapi_base.api_call_counter_decorator
    def calculate_coverage(self, coverage_type: int, agent_id: str = None):
        """Calculate coverage of different resources in the test case set.

        Args:
          coverage_type: The type of coverage requested.
            INTENT = 1
            PAGE_TRANSITION = 2
            TRANSITION_ROUTE_GROUP = 3
          agent: The CX agent to calculate coverage for.
            `projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>`

        Returns:
          The Coverage response of the test cases for the type specified.
        """

        if not agent_id:
            agent_id = self.agent_id

        if coverage_type not in [1, 2, 3]:
            raise ValueError(
                f"Invalid coverage_type: {coverage_type}. coverage_type must "
                "be must be 1, 2 or 3"
            )

        request = types.test_case.CalculateCoverageRequest()
        request.agent = agent_id
        request.type_ = coverage_type

        client_options = self._set_region(agent_id)
        client = services.test_cases.TestCasesClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.calculate_coverage(request)
        return response

    def get_test_case_results_df(self, agent_id=None, retest_all=False):
        """Convert Test Cases to Dataframe.

        Gets the test case results for this agent, and generates a dataframe
        with their details. Any tests without a result will be run in a batch.

        Args:
          agent_id: The agent to create the test case for. Format:
              `projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>`
          retest_all: if true, all test cases are re-run,
            regardless of whether or not they had a result

        Returns:
          DataFrame of test case results for this agent, with columns:
            display_name, id, short_id, tags, creation_time, start_flow,
            start_page, passed, test_time
        """
        if agent_id:
            self.agent_id = agent_id

        dfcx_flows = flows.Flows(creds=self.creds, agent_id=self.agent_id)
        dfcx_pages = pages.Pages(creds=self.creds)
        flows_map = dfcx_flows.get_flows_map(agent_id=self.agent_id)
        pages_map = {}
        for flow_id in flows_map.keys():
            pages_map[flow_id] = dfcx_pages.get_pages_map(flow_id=flow_id)

        test_case_results = self.list_test_cases(self.agent_id)
        retest_ids = []
        test_case_rows = []

        for test_case in test_case_results:
            row = self._process_test_case(test_case, flows_map, pages_map)
            test_case_rows.append(row)
            test_result = self._convert_test_result_to_string(test_case)
            if retest_all or test_result == "TEST_RESULT_UNSPECIFIED":
                retest_ids.append(test_case.name)

        # Create dataframe
        test_case_df = pd.concat(test_case_rows)

        # Retest any that haven't been run yet
        if len(retest_ids) > 0:
            test_case_df = self._retest_cases(test_case_df,retest_ids)

        return test_case_df



"""Flow Resource functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
from typing import Dict, List
from google.cloud.dialogflowcx_v3beta1 import services
from google.cloud.dialogflowcx_v3beta1 import types
from google.protobuf import field_mask_pb2
from dfcx_scrapi.core import scrapi_base
from dfcx_scrapi.core import pages

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

logger = logging.getLogger(__name__)


class Flows(scrapi_base.ScrapiBase):
    """Core Class for CX Flow Resource functions."""

    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        scope=False,
        flow_id: str = None,
        agent_id: str = None,
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

        if flow_id:
            self.flow_id = flow_id

        self.agent_id = agent_id
        self.pages = pages.Pages(creds=self.creds)

    # TODO: Migrate to Flow Builder class when ready
    @staticmethod
    def _build_nlu_settings(
        model_type: str = "STANDARD",
        classification_threshold: float = 0.3,
        model_training_mode: str = "MANUAL",
    ):
        """Builds the NLU Settings object to be used with Flow objects.

        Args:
          model_type: ONEOF `STANDARD`, `ADVANCED`, `CUSTOM`. Defaults to
            `STANDARD`.
          classification_threshold: To filter out false positive results and
            still get variety in matched natural language inputs for your
            agent, you can tune the machine learning classification threshold.
            If the returned score value is less than the threshold value, then
            a no-match event will be triggered. The score values range from 0.0
            (completely uncertain) to 1.0 (completely certain). If set to 0.0,
            the default of 0.3 is used.
          model_training_mode: ONEOF `AUTOMATIC`, `MANUAL`. Defaults to
            `MANUAL`
        """
        model_type_map = {"STANDARD": 1, "CUSTOM": 2, "ADVANCED": 3}

        model_training_map = {"AUTOMATIC": 1, "MANUAL": 2}

        nlu_settings = types.NluSettings()
        nlu_settings.classification_threshold = classification_threshold

        if model_type in model_type_map:
            nlu_settings.model_type = model_type_map[model_type]
        else:
            raise KeyError(
                f"`{model_type}` is invalid. `model_type` must be "
                "one of `STANDARD`, `ADVANCED`, `CUSTOM`."
            )

        if model_training_mode in model_training_map:
            nlu_settings.model_training_mode = model_training_map[
                model_training_mode
            ]
        else:
            raise KeyError(
                f"`{model_training_mode}` is invalid. "
                "`model_training_mode` must be one of `AUTOMATIC`, `MANUAL`."
            )

        return nlu_settings

    def get_flows_map(self, agent_id: str, reverse=False):
        """Exports Agent Flow Names and UUIDs into a user friendly dict.

        Args:
          agent_id: the formatted CX Agent ID to use
          reverse: (Optional) Boolean flag to swap key:value -> value:key

        Returns:
          Dictionary containing flow UUIDs as keys and display names as values
        """

        if reverse:
            flows_dict = {
                flow.display_name: flow.name
                for flow in self.list_flows(agent_id=agent_id)
            }

        else:
            flows_dict = {
                flow.name: flow.display_name
                for flow in self.list_flows(agent_id=agent_id)
            }

        return flows_dict

    def get_flow_page_map(
            self, agent_id: str, rate_limit: float = 1.0
            ) -> Dict[str, Dict[str, str]]:
        """Exports a user friendly dict containing Flows, Pages, and IDs
        This method builds on top of `get_flows_map` and builds out a nested
        dictionary containing all of the Page Display Names and UUIDs contained
        within each Flow. Output Format:
          {
            <FLOW_DISPLAY_NAME>: {
                'id': <FLOW_UUID>
                'pages': { <PAGE_DISPLAY_NAME> : <PAGE_UUID> }
            }
          }

        Args:
          agent_id: the formatted CX Agent ID to use

        Returns:
          Dictionary containing Flow Names/UUIDs and Page Names/UUIDs
        """
        flow_page_map = {}

        flows_map = self.get_flows_map(agent_id, reverse=True)

        for flow in flows_map:
            pages_map = self.pages.get_pages_map(
                flows_map[flow], reverse=True)
            flow_page_map[flow] = {"id": flows_map[flow], "pages": pages_map}
            time.sleep(rate_limit)

        return flow_page_map

    @scrapi_base.api_call_counter_decorator
    def train_flow(self, flow_id: str) -> str:
        """Trains the specified flow.

        Args:
          flow_id: CX flow ID string in the following format
            projects/<PROJECT ID>/locations/<LOCATION ID>/agents/<AGENT ID>/
              flows/<FLOW ID>

        Returns:
          A Long Running Operation (LRO) ID that can be used to
            check the status of the export using
              dfcx_scrapi.core.operations->get_lro()
        """

        request = types.flow.TrainFlowRequest()
        request.name = flow_id

        client_options = self._set_region(flow_id)
        client = services.flows.FlowsClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.train_flow(request)

        return response

    @scrapi_base.api_call_counter_decorator
    def list_flows(self, agent_id: str) -> List[types.Flow]:
        """Get a List of all Flows in the current Agent.

        Args:
          agent_id: CX Agent ID string in the proper format
            projects/<PROJECT ID>/locations/<LOCATION ID>/agents/<AGENT ID>

        Returns:
          List of Flow objects
        """

        request = types.flow.ListFlowsRequest()
        request.parent = agent_id

        client_options = self._set_region(agent_id)
        client = services.flows.FlowsClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.list_flows(request)

        flows = []
        for page in response.pages:
            for flow in page.flows:
                flows.append(flow)
        return flows

    def get_flow_by_display_name(
        self, display_name: str, agent_id: str
    ) -> types.Flow:
        """Get a single CX Flow object based on its display name.

        Args:
          display_name: The display name of the desired Flow.
          agent_id: CX Agent ID in which the flow exists.

        Returns:
          A single CX Flow object
        """

        flows_map = self.get_flows_map(agent_id=agent_id, reverse=True)

        if display_name in flows_map:
            flow_id = flows_map[display_name]
        else:
            raise ValueError(
                f'Flow "{display_name}" '
                f"does not exist in the specified agent."
            )

        flow = self.get_flow(flow_id=flow_id)

        return flow

    @scrapi_base.api_call_counter_decorator
    def get_flow(self, flow_id: str) -> types.Flow:
        """Get a single CX Flow object.

        Args:
          flow_id: CX Flow ID in the proper format

        Returns:
          A single CX Flow object
        """

        client_options = self._set_region(flow_id)
        client = services.flows.FlowsClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.get_flow(name=flow_id)

        return response

    @scrapi_base.api_call_counter_decorator
    def create_flow(
        self,
        agent_id: str,
        display_name: str = None,
        language_code: str = "en",
        obj: types.Flow = None,
        **kwargs,
    ):
        """Create a Dialogflow CX Flow with given display name.

        If the user provides an existing Flow object, a new CX Flow will be
        created based on this object and any other input/kwargs will be
        discarded.

        Args:
          agent_id: DFCX Agent id where the Flow will be created
          display_name: Human readable display name for the CX Flow
          obj: (Optional) Flow object to create in proto format

        Returns:
          The newly created CX Flow resource object.
        """
        request = types.flow.CreateFlowRequest()
        request.parent = agent_id
        request.language_code = language_code

        if obj:
            flow_obj = obj
            request.flow = flow_obj

        else:
            flow_obj = types.Flow()
            flow_obj.display_name = display_name

            # set optional args as agent attributes
            for key, value in kwargs.items():
                setattr(flow_obj, key, value)

            request.flow = flow_obj

        client_options = self._set_region(agent_id)
        client = services.flows.FlowsClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.create_flow(request)

        return response

    @scrapi_base.api_call_counter_decorator
    def update_flow(
        self, flow_id: str, obj: types.Flow = None, **kwargs
    ) -> types.Flow:
        """Update a single specific CX Flow object.

        Args:
          flow_id: CX Flow ID in the proper format
          obj: (Optional) a single CX Flow object of types.Flow

        Returns:
          A copy of the updated Flow object
        """

        if obj:
            flow = obj
            flow.name = flow_id
        else:
            flow = self.get_flow(flow_id)

        # set flow attributes to args
        for key, value in kwargs.items():
            setattr(flow, key, value)
        paths = kwargs.keys()
        mask = field_mask_pb2.FieldMask(paths=paths)

        client_options = self._set_region(flow_id)
        client = services.flows.FlowsClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.update_flow(flow=flow, update_mask=mask)

        return response

    def update_nlu_settings(self, flow_id: str, **kwargs):
        """Updates flow to new NLU setting.

        Args:
          flow_id: flow id to update nlu settings for.
            model_type: (Optional) [0:unspecified, 1:MODEL_TYPE_STANDARD,
              2:Custom, 3:Advanced]
            classification_threshold: (Optional) threshold for the flow
            model_training_mode: (Optional) [0:unspecified, 1:automatic,
              2:'manual]
        """

        flow = self.get_flow(flow_id)
        current_settings = flow.nlu_settings
        for key, value in kwargs.items():
            setattr(current_settings, key, value)
        self.update_flow(flow_id=flow_id, nlu_settings=current_settings)

    @scrapi_base.api_call_counter_decorator
    def export_flow(
        self, flow_id: str, gcs_path: str, ref_flows: bool = True
    ) -> Dict[str, str]:
        """Exports DFCX Flow(s) into GCS bucket.

        Args:
          flow_id: the formatted CX Flow ID to export
          gcs_path: The `Google Cloud Storage URI to export the flow to. The
            format of this URI must be ``gs://<bucket-name>/<object-name>``. If
            left unspecified, the serialized flow is returned inline.
          ref_flows: Whether to export flows referenced by the specified flow.

        Returns:
          A Long Running Operation result. If successful the LRO result will
            return the Google Cloud Storage URI from the Export Flow request.
            Otherwise, it will return the corresponding error.
        """
        request = types.flow.ExportFlowRequest()
        request.name = flow_id
        request.include_referenced_flows = ref_flows
        request.flow_uri = gcs_path

        client_options = self._set_region(flow_id)
        client = services.flows.FlowsClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.export_flow(request)

        return response.result()

    @scrapi_base.api_call_counter_decorator
    def export_flow_inline(self, flow_id: str, ref_flows: bool = True) -> bytes:
        """Export a Flow, returning uncompressed raw byte content for flow.

        Args:
          flow_id: the formatted CX Flow ID to export
          ref_flows: Whether to export flows referenced by the specified flow.

        Returns:
          Bytes representing the content of the flow.
        """
        request = types.flow.ExportFlowRequest()
        request.name = flow_id
        request.include_referenced_flows = ref_flows

        client_options = self._set_region(flow_id)
        client = services.flows.FlowsClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.export_flow(request)

        return (response.result()).flow_content

    @scrapi_base.api_call_counter_decorator
    def import_flow(
        self,
        agent_id: str,
        gcs_path: str = None,
        flow_content: bytes = None,
        import_option: str = "KEEP",
    ) -> Dict[str, str]:
        """Imports a DFCX Flow to CX Agent. Flow can be imported from a
        GCS bucket or from raw bytes.

        Args:
          agent_id: the CX Agent ID to import the flow into.
          gcs_path: the `Google Cloud Storage URI to import flow from.
            the format of this URI must be ``gs://<bucket-name>/<object-name>``.
          flow_content: uncompressed raw byte content for flow.
          import_option: one of 'FALLBACK' or 'KEEP'. Defaults to 'KEEP'

        Returns:
          A Long Running Operation result. If successful the LRO result will
            return the Flow ID of the newly imported Flow.
            Otherwise, it will return the corresponding error.
        """

        if gcs_path and flow_content:
            raise ValueError(
                "gcs_path or flow_content required (But not both!)."
            )

        if not gcs_path and not flow_content:
            raise ValueError(
                "gcs_path or flow_content required (But not both!)."
            )

        request = types.flow.ImportFlowRequest()
        request.parent = agent_id
        request.flow_uri = gcs_path
        request.flow_content = flow_content
        request.import_option = import_option

        client_options = self._set_region(agent_id)
        client = services.flows.FlowsClient(
            credentials=self.creds, client_options=client_options
        )

        response = client.import_flow(request)

        return response

    @scrapi_base.api_call_counter_decorator
    def delete_flow(self, flow_id: str, force: bool = False):
        """Deletes a single CX Flow Object resource.

        Args:
          flow_id: flow to delete
          force: False means a flow will not be deleted if a route to the flow
            exists, True means the flow will be deleted as well as all the
            transition routes leading to the flow.
        """

        request = types.DeleteFlowRequest()
        request.name = flow_id
        request.force = force

        client_options = self._set_region(flow_id)
        client = services.flows.FlowsClient(
            credentials=self.creds, client_options=client_options
        )

        client.delete_flow(request)



"""Webhook processing methods and functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

from dfcx_scrapi.agent_extract import common
from dfcx_scrapi.agent_extract import types

class Webhooks:
    """Webhook linter methods and functions."""

    def __init__(self):
        self.common = common.Common()

    @staticmethod
    def build_webhook_path_list(agent_local_path: str):
        """Builds a list of webhook file locations."""
        root_dir = agent_local_path + "/webhooks"

        webhook_paths = []

        for webhook_file in os.listdir(root_dir):
            webhook_file_path = f"{root_dir}/{webhook_file}"
            webhook_paths.append(webhook_file_path)

        return webhook_paths

    @staticmethod
    def get_service_type(webhook: types.Webhook) -> str:
        """Get the type of Webhook Service that is cofigured."""
        if "genericWebService" in webhook.data:
            webhook.service_type = "Generic Web Service"

        else:
            webhook.service_type = "Other"

        return webhook.service_type

    def process_webhook(self, webhook: types.Webhook, stats: types.AgentData
            ) -> types.AgentData:
        """Process a single Webhook file."""

        with open(webhook.dir_path, "r", encoding="UTF-8") as webhook_file:
            webhook.data = json.load(webhook_file)
            webhook.resource_id = webhook.data.get("name", None)
            webhook.display_name = webhook.data.get("displayName", None)
            webhook.service_type = self.get_service_type(webhook)

            timeout_dict = webhook.data.get("timeout", None)
            if timeout_dict:
                webhook.timeout = timeout_dict.get("seconds", None)

            webhook_file.close()

        full_webhook_id = f"{stats.agent_id}/webhooks/{webhook.resource_id}"
        webhook.data["name"] = full_webhook_id
        stats.webhooks.append(webhook.data)
        stats.total_webhooks += 1

        return stats

    def process_webhooks_directory(self, agent_local_path: str,
                                   stats: types.AgentData) -> types.AgentData:
        """Processing the top level Webhooks Dir in the JSON Package structure.

        The following files exist under the `webhooks` dir:
        - <webhook-name>.json
        """
        # Create a list of all Webhook paths to iter through
        webhook_paths = self.build_webhook_path_list(agent_local_path)

        for webhook_path in webhook_paths:
            webhook = types.Webhook()
            webhook.dir_path = webhook_path

            stats = self.process_webhook(webhook, stats)

            full_webhook_id = f"{stats.agent_id}/webhooks/{webhook.resource_id}"
            stats.webhooks_map[webhook.display_name] = full_webhook_id

        return stats



"""Route Groups processing methods and functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json

from dfcx_scrapi.agent_extract import common
from dfcx_scrapi.agent_extract import types
from dfcx_scrapi.agent_extract import routes


class RouteGroups:
    """Route Groups processing methods and functions."""

    def __init__(self):
        self.special_pages = [
            "End Session",
            "End Flow",
            "Start Page",
            "Current Page",
            "Previous Page",
        ]

        self.common = common.Common()
        self.routes = routes.Fulfillments()

    @staticmethod
    def build_route_group_path_list(flow_local_path: str):
        """Builds a list of files, each representing a Route Group.

        Ex: /path/to/agent/flows/<flow_dir>/transitionRouteGroups/<rg.json>
        """
        root_dir = flow_local_path + "/transitionRouteGroups"

        if "transitionRouteGroups" in os.listdir(flow_local_path):
            rg_paths = []

            for rg_file in os.listdir(root_dir):
                rg_file_path = f"{root_dir}/{rg_file}"
                rg_paths.append(rg_file_path)

        return rg_paths

    def process_route_group(self, rg: types.RouteGroup, stats: types.AgentData):
        """Process a single Route Group."""
        rg.display_name = self.common.parse_filepath(rg.rg_file, "route_group")
        rg.display_name = self.common.clean_display_name(rg.display_name)

        with open(rg.rg_file, "r", encoding="UTF-8") as route_group_file:
            rg.data = json.load(route_group_file)
            rg.resource_id = rg.data.get("name", None)
            rg.display_name = rg.data.get("displayName", None)
            rg.routes = rg.data.get("transitionRoutes", None)

            stats = self.routes.process_routes(rg, stats)

            route_group_file.close()

        full_flow_id = f"{stats.agent_id}/flows/{rg.flow.resource_id}"
        full_rg_id = f"{full_flow_id}/transitionRouteGroups/{rg.resource_id}"
        stats.route_groups_map[
            rg.flow.display_name]["route_groups"][rg.display_name] = full_rg_id
        stats.route_groups[rg.flow.display_name].append(rg.data)

        return stats

    def process_route_groups_directory(
            self, flow: types.Flow, stats: types.AgentData):
        """Process Route Groups dir in the JSON Package structure."""
        if "transitionRouteGroups" in os.listdir(flow.dir_path):
            # Create a list of all Route Group paths to iter through
            rg_paths = self.build_route_group_path_list(flow.dir_path)
            stats.total_route_groups += len(rg_paths)

            full_flow_id = f"{stats.agent_id}/flows/{flow.resource_id}"
            stats.route_groups_map[flow.display_name] = {
                "id": full_flow_id,
                "route_groups": {}
            }
            stats.route_groups[flow.display_name] = []

            for rg_path in rg_paths:
                rg = types.RouteGroup(flow=flow)
                rg.rg_file = rg_path
                stats = self.process_route_group(rg, stats)

        return stats
